"""Schemas for various metadata objects.

This module contains schemas (as TypedDict) for metadata from various
PLAYA objects.

"""

import logging
from typing import Any, Dict, List, Set, Tuple, Union

try:
    # We only absolutely need this when using Pydantic TypeAdapter
    from typing_extensions import TypedDict
except ImportError:
    from typing import TypedDict

from playa.data._asobj import asobj
from playa.document import Destinations as _Destinations
from playa.document import Document as _Document
from playa.font import Font as _Font
from playa.fontmetrics import FONT_METRICS
from playa.outline import (
    Action as _Action,
    Destination as _Destination,
    DestinationXYZ,
    DestinationFitH,
    DestinationFitV,
    DestinationFitR,
)
from playa.outline import Tree as _OutlineTree, Item as _OutlineItem
from playa.page import Annotation as _Annotation
from playa.page import Page as _Page
from playa.parser import IndirectObject as _IndirectObject
from playa.pdftypes import MATRIX_IDENTITY
from playa.pdftypes import ContentStream as _ContentStream
from playa.pdftypes import (
    PDFObject,
    PSLiteral,
    Point,
    dict_value,
    int_value,
    matrix_value,
    rect_value,
    resolve1,
    stream_value,
)
from playa.structure import ContentItem as _StructContentItem
from playa.structure import ContentObject as _StructContentObject
from playa.structure import Element as _Element
from playa.structure import Tree as _Tree
from playa.utils import Matrix, Rect

log = logging.getLogger(__name__)


class Document(TypedDict, total=False):
    """Metadata for a PDF document."""

    pdf_version: str
    """Version of the PDF standard this document implements."""
    is_printable: bool
    """Should the user be allowed to print?"""
    is_modifiable: bool
    """Should the user be allowed to modify?"""
    is_extractable: bool
    """Should the user be allowed to extract text?"""
    encryption: "Encryption"
    """Encryption information for this document."""
    outlines: List["Outline"]
    """Outline for this document."""
    destinations: Dict[str, "Destination"]
    """Named destinations for this document."""
    structure: "StructTree"
    """Logical structure for this document.."""
    pages: List["Page"]
    """Pages in this document."""
    objects: List["IndirectObject"]
    """Indirect objects in this document."""


class Encryption(TypedDict, total=False):
    """Encryption information."""

    ids: Tuple[str, str]
    """ID values for encryption."""
    encrypt: dict
    """Encryption properties."""


class Outline(TypedDict, total=False):
    """Outline item for a PDF document."""

    title: str
    """Title of this outline entry."""
    destination: "Destination"
    """Destination (or target of GoTo action)."""
    children: List["Outline"]
    """Children of this entry."""


class Destination(TypedDict, total=False):
    """Destination for an outline entry or annotation."""

    page_idx: int
    """Zero-based index of destination page."""
    display: str
    """How to display the destination on that page."""
    pos: Point
    """Target position."""
    top: float
    """Target top edge."""
    left: float
    """Target left edge."""
    bbox: Rect
    """Target zoom box."""
    zoom: float
    """Zoom factor."""


class Action(TypedDict, total=False):
    """Action for an outline entry or annotation."""

    type: str
    """Type of action."""
    destination: Destination
    """Destination for a GoTo action."""
    uri: str
    """URI for a URI action."""


class StructContentObject(TypedDict, total=False):
    page_idx: int
    """Page index of content object."""
    props: dict
    """Properties of content object."""


class StructContentItem(TypedDict, total=False):
    page_idx: int
    """Page index of content object."""
    mcid: int
    """Marked content section ID."""
    stream: "StreamObject"
    """Content stream, if any."""


class StructElement(TypedDict, total=False):
    """Node in logical structure tree."""

    type: str
    """Type of structure element (or "StructTreeRoot" for root)."""
    role: str
    """Role of structure element (root has no role)."""
    page_idx: int
    """Page on which this structure element's content begins."""
    title: str
    """Title of structure element."""
    language: str
    """Language of structure element."""
    alternate_description: str
    """Alternate description."""
    abbreviation_expansion: str
    """Abbreviation expansion."""
    actual_text: str
    """Unicode text content."""
    children: List["StructElement"]
    """Children of this node."""
    attributes: dict
    """Structure attributes."""
    class_name: str
    """Structure attribute class name."""


class StructTree(TypedDict, total=False):
    """Logical structure tree for a PDF document."""

    root: StructElement
    """Root node of the tree."""


class Page(TypedDict, total=False):
    """Metadata for a PDF page."""

    page_idx: int
    """0-based page number."""
    page_label: Union[str, None]
    """Page label (could be roman numerals, letters, etc)."""
    page_id: int
    """Indirect object ID."""
    mediabox: Rect
    """Extent of physical page, in base units (1/72 inch)."""
    cropbox: Rect
    """Extent of visible area, in base units (1/72 inch)."""
    rotate: int
    """Page rotation in degrees."""
    resources: "Resources"
    """Page resources."""
    annotations: List["Annotation"]
    """Page annotations."""
    contents: List["StreamObject"]
    """Metadata for content streams."""


class Annotation(TypedDict, total=False):
    type: str
    """Type of annotation."""
    bbox: Rect
    """Bounding rectangle for annotation."""
    contents: str
    """Text contents."""
    name: str
    """Annotation name, uniquely identifying this annotation."""
    mtime: str
    """String describing date and time when annotation was most recently
    modified."""
    destination: Destination
    """Destination of a link annotation."""
    action: Action
    """Action for a link annotation."""


class XObject(TypedDict, total=False):
    type: str
    """Type of XObject."""
    xobject_id: int
    """Indirect object ID."""
    genno: int
    """Generation number."""
    length: int
    """Length of raw stream data."""
    filters: List[str]
    """List of filters."""
    params: List[dict]
    """Filter parameters."""
    resources: "Resources"
    """Resources specific to this XObject, if any."""


class StreamObject(TypedDict, total=False):
    stream_id: int
    """Indirect object ID."""
    genno: int
    """Generation number."""
    length: int
    """Length of raw stream data."""
    filters: List[str]
    """List of filters."""
    params: List[dict]
    """Filter parameters."""


class IndirectObject(TypedDict, total=False):
    objid: int
    """Indirect object ID."""
    genno: int
    """Generation number."""
    type: str
    """Name of Python type to which this object was converted."""
    obj: Union[float, int, str, bool, dict, list]
    """Object metadata (for streams) or data (otherwise)."""


class Font(TypedDict, total=False):
    name: str
    """Font name."""
    type: str
    """Font type (Type1, Type0, TrueType, Type3, etc)."""
    ascent: float
    """Ascent in glyph space units."""
    descent: float
    """Descent in glyph space units."""
    italic_angle: float
    """Italic angle."""
    default_width: float
    """Default character width in glyph space units."""
    leading: float
    """Leading in glyph space units."""
    cap_height: float
    """Top of capital letters, in glyph space units."""
    xheight: float
    """Top of flat nonascending lowercase letters, in glyph space units."""
    stem_v: float
    """Thickness of dominant vertical stems."""
    stem_h: float
    """Thickness of dominant horizontal stems."""
    avg_width: float
    """Average width of glyphs in glyph space units."""
    max_width: float
    """Maximum width of glyphs in glyph space units."""
    stretch: str
    """Font stretch value (e.g. Condensed, Expanded)."""
    weight: float
    """Numeric weight value (100, 200, 300... 900)"""
    flags: List[str]
    """Set of flags (e.g. FixedPitch, Script, etc) for this font."""
    bbox: Rect
    """Bounding box in glyph space units."""
    matrix: Matrix
    """Matrix mapping glyph space to text space (Type3 fonts only)."""
    cidfont: "Font"
    """Descendant CIDFont (Type0 fonts only)."""


FONT_FLAGS = {
    "FixedPitch": 1,
    "Serif": 2,
    "Symbolic": 3,
    "Script": 4,
    "Nonsymbolic": 6,
    "Italic": 7,
    "AllCap": 17,
    "SmallCap": 18,
    "ForceBold": 19,
}


def flags_to_list(flags: int) -> List[str]:
    return [k for k, v in FONT_FLAGS.items() if flags & (1 << v)]


FONT_ATTRS = {
    "ascent": "Ascent",
    "descent": "Descent",
    "italic_angle": "ItalicAngle",
    "default_width": "MissingWidth",
    "leading": "Leading",
    "cap_height": "CapHeight",
    "xheight": "XHeight",
    "stem_v": "StemV",
    "stem_h": "StemH",
    "avg_width": "AvgWidth",
    "max_width": "MaxWidth",
    "stretch": "FontStretch",
    "weight": "FontWeight",
}


def font_from_spec(spec: Dict[str, Any]) -> Font:
    basefont = asobj(resolve1(spec.get("BaseFont")))
    if basefont is None:
        basefont = asobj(resolve1(spec.get("Name")))
    font = Font(
        name=basefont,
        type=asobj(resolve1(spec.get("Subtype"))),
    )
    desc: PDFObject
    if basefont in FONT_METRICS:
        desc, _ = FONT_METRICS[basefont]
    else:
        desc = resolve1(spec.get("FontDescriptor"))
    if desc is not None:
        desc = dict_value(desc)
        for attr in (
            "ascent",
            "descent",
            "italic_angle",
            "default_width",
            "leading",
            "cap_height",
            "xheight",
            "stem_v",
            "stem_h",
            "avg_width",
            "max_width",
            "stretch",
            "weight",
        ):
            key = FONT_ATTRS[attr]
            val = desc.get(key)
            if val:
                font[attr] = asobj(val)
        bbox = rect_value(desc.get("FontBBox", (0, 0, 0, 0)))
        if bbox != (0, 0, 0, 0):
            font["bbox"] = bbox
        flags = desc.get("Flags", 0)
        if flags:
            flaglist = flags_to_list(int_value(flags))
            if flaglist:
                font["flags"] = flaglist
        if font["name"] is None and "FontName" in desc:
            font["name"] = asobj(resolve1(desc["FontName"]))
    sub = resolve1(spec.get("DescendantFonts"))
    if sub and isinstance(sub, list) and sub[0] is not None:
        font["cidfont"] = font_from_spec(dict_value(sub[0]))
    # For Type3 fonts
    if "bbox" not in font:
        bbox = spec.get("FontBBox", (0, 0, 0, 0))
        if bbox != (0, 0, 0, 0):
            font["bbox"] = bbox
    if "FontMatrix" in spec:
        font["matrix"] = matrix_value(spec["FontMatrix"])
    return font


class Resources(TypedDict, total=False):
    ext_gstates: Dict[str, dict]
    """Extended graphic state dictionaries."""
    color_spaces: Dict[str, Any]
    """Color space descriptors."""
    patterns: Dict[str, Any]
    """Pattern objects."""
    shadings: Dict[str, dict]
    """Shading dictionaries."""
    xobjects: Dict[str, "XObject"]
    """XObject streams."""
    fonts: Dict[str, Font]
    """Font dictionaries."""
    procsets: List[str]
    """Procedure set names."""
    properties: Dict[str, dict]
    """property dictionaires."""


def xobject_from_stream(
    obj: _ContentStream, seen: Union[Set[int], None] = None
) -> XObject:
    if seen is None:
        seen = set()
    stream = stream_metadata(obj)
    seen.add(stream["stream_id"])
    xobj = XObject(
        xobject_id=stream["stream_id"],
        genno=stream["genno"],
        length=stream["length"],
        type=asobj(obj.attrs["Subtype"]),
    )
    if "filters" in stream:
        xobj["filters"] = stream["filters"]
    if "params" in stream:
        xobj["params"] = stream["params"]
    resources = resolve1(obj.attrs.get("Resources"))
    if resources:
        xobj["resources"] = resources_from_dict(dict_value(resources), seen)
    return xobj


RESOURCE_DICTS = {
    "ext_gstates": "ExtGState",
    "color_spaces": "ColorSpace",
    "patterns": "Pattern",
    "shadings": "Shading",
    "properties": "Properties",
}


def resources_from_dict(
    resources: Dict[str, Any], seen: Union[Set[int], None] = None
) -> Resources:
    if seen is None:
        seen = set()
    res = Resources()
    for attr in "ext_gstates", "color_spaces", "patterns", "shadings", "properties":
        key = RESOURCE_DICTS[attr]
        d = resolve1(resources.get(key))
        if d and isinstance(d, dict):
            res[attr] = {k: asobj(resolve1(v)) for k, v in d.items()}
    d = resolve1(resources.get("XObject"))
    if d and isinstance(d, dict):
        xobjects = {}
        for k, v in d.items():
            try:
                stream = stream_value(v)
                # Use a reference if we've already seen it
                if stream.objid in seen:
                    xobjects[k] = asobj(v)
                    continue
                if stream.objid is not None:
                    seen.add(stream.objid)
                xobjects[k] = xobject_from_stream(stream, seen)
            except Exception as e:
                log.warning("Failed to get XObject %s: %s", k, e)
        if xobjects:
            res["xobjects"] = xobjects
    p = resolve1(resources.get("ProcSet"))
    if p and isinstance(p, list):
        res["procsets"] = asobj(p)
    d = resolve1(resources.get("Font"))
    if d and isinstance(d, dict):
        fonts = {}
        for k, v in d.items():
            spec = resolve1(v)
            if not isinstance(spec, dict):
                log.warning("Invalid font specification: %r", spec)
            else:
                fonts[k] = font_from_spec(spec)
        if fonts:
            res["fonts"] = fonts
    return res


def stream_metadata(obj: _ContentStream) -> StreamObject:
    # These really cannot be None!
    assert obj.objid is not None
    assert obj.genno is not None
    length = int_value(obj.attrs["Length"])
    cs = StreamObject(stream_id=obj.objid, genno=obj.genno, length=length)
    fps = obj.get_filters()
    if fps:
        filters, params = zip(*fps)
        if any(filters):
            cs["filters"] = asobj(filters)
        if any(params):
            cs["params"] = asobj(params)
    return cs


@asobj.register
def asobj_page(obj: _Page) -> Page:
    page = Page(
        page_idx=obj.page_idx,
        page_label=obj.label,
        page_id=obj.pageid,
        mediabox=obj.mediabox,
        cropbox=obj.cropbox,
        rotate=obj.rotate,
        resources=resources_from_dict(obj.resources),
        annotations=[asobj(annot) for annot in obj.annotations],
    )
    contents = []
    for stream in obj.streams:
        try:
            contents.append(stream_metadata(stream))
        except Exception as e:
            log.warning("Failed to get stream metadata from %r: %s", stream, e)
    if contents:
        page["contents"] = contents
    return page


@asobj.register
def asobj_annotation(obj: _Annotation) -> Annotation:
    annot = Annotation(type=obj.type, bbox=obj.bbox)
    for attr in "contents", "name", "mtime":
        val = getattr(obj, attr)
        if val is not None:
            annot[attr] = asobj(val)
    dest = resolve1(obj.props.get("Dest"))
    if isinstance(dest, (PSLiteral, bytes, list)):
        annot["destination"] = asobj(_Destination.from_dest(dest, obj.page.doc))
    if "A" in obj.props:
        annot["action"] = asobj(_Action.from_dict(obj.props["A"], obj.page.doc))
    return annot


# Note this is not actually used for font resources, since for the
# moment the Font class we inherited from pdfminer.six lacks various
# metadata.  It's here anyway in case you call asobj() on a Font.
@asobj.register
def asobj_font(obj: _Font) -> Font:
    font_type = obj.__class__.__name__
    if font_type != "CIDFont":
        font_type = font_type.replace("Font", "")
    font = Font(
        name=obj.fontname,
        type=font_type,
    )
    for attr in (
        "ascent",
        "descent",
        "italic_angle",
        "default_width",
        "leading",
    ):
        val = getattr(obj, attr, None)
        if val:
            font[attr] = asobj(val)
    if obj.bbox != (0, 0, 0, 0):
        font["bbox"] = obj.bbox
    if hasattr(obj, "matrix") and obj.matrix != MATRIX_IDENTITY:
        font["matrix"] = obj.matrix
    return font


@asobj.register
def asobj_obj(obj: _IndirectObject) -> IndirectObject:
    return IndirectObject(
        objid=obj.objid,
        genno=obj.genno,
        type=type(obj.obj).__name__,
        obj=asobj(obj.obj),
    )


@asobj.register
def asobj_stream(obj: _ContentStream) -> Dict:
    meta = asobj(obj.attrs)
    meta["stream_id"] = obj.objid
    return meta


@asobj.register
def asobj_outline_tree(obj: _OutlineTree) -> List[Outline]:
    return [asobj(outline) for outline in obj]


@asobj.register
def asobj_outline_item(obj: _OutlineItem, recurse: bool = True) -> Outline:
    out = Outline()
    out["title"] = obj.title
    if obj.destination is not None:
        out["destination"] = asobj(obj.destination)
    if recurse:
        children = list(obj)
        if children:
            out["children"] = asobj(children)
    return out


@asobj.register
def asobj_destinations(obj: _Destinations) -> Dict[str, Destination]:
    return {name: asobj(dest) for name, dest in obj.items()}


@asobj.register
def asobj_destination(obj: _Destination) -> Destination:
    dest = Destination()
    if obj.page_idx is not None:
        dest["page_idx"] = obj.page_idx
    if obj.display is not None:
        dest["display"] = asobj(obj.display)
    return dest


@asobj.register
def asobj_destination_xyz(obj: DestinationXYZ) -> Destination:
    dest = asobj_destination(obj)
    if obj.zoom is not None:
        dest["zoom"] = obj.zoom
    dest["pos"] = asobj(obj.pos)
    return dest


@asobj.register
def asobj_destination_fith(obj: DestinationFitH) -> Destination:
    dest = asobj_destination(obj)
    if obj.top is not None:
        dest["top"] = obj.top
    return dest


@asobj.register
def asobj_destination_fitv(obj: DestinationFitV) -> Destination:
    dest = asobj_destination(obj)
    if obj.left is not None:
        dest["left"] = obj.left
    return dest


@asobj.register
def asobj_destination_fitr(obj: DestinationFitR) -> Destination:
    dest = asobj_destination(obj)
    dest["bbox"] = obj.bbox
    return dest


@asobj.register
def asobj_action(obj: _Action) -> Action:
    action = Action(type=asobj(obj.type))
    if "URI" in obj.props:
        action["uri"] = asobj(obj.props["URI"])
    if "D" in obj.props:
        dest = resolve1(obj.props["D"])
        if isinstance(dest, (PSLiteral, bytes, list)):
            action["destination"] = asobj(_Destination.from_dest(dest, obj.doc))
    return action


@asobj.register
def asobj_content_item(obj: _StructContentItem) -> StructContentItem:
    item = StructContentItem(mcid=obj.mcid)
    page = obj.page
    if page is not None:
        item["page_idx"] = page.page_idx
    if obj.stream is not None:
        item["stream"] = stream_metadata(obj.stream)
    return item


@asobj.register
def asobj_content_object(obj: _StructContentObject) -> StructContentObject:
    item = StructContentObject(props=asobj(obj.props))
    page = obj.page
    if page is not None:
        item["page_idx"] = page.page_idx
    return item


@asobj.register
def asobj_structelement(obj: _Element, recurse: bool = True) -> StructElement:
    el = StructElement(type=obj.type)
    page = obj.page
    for attr in (
        "role",
        "title",
        "language",
        "alternate_description",
        "abbreviation_expansion",
        "actual_text",
        "attributes",
        "class_name",
    ):
        val = getattr(obj, attr)
        if val is not None:
            el[attr] = asobj(val)
    if page is not None:
        el["page_idx"] = page.page_idx
    if recurse:
        children = [asobj(el) for el in obj]
        if children:
            el["children"] = children
    return el


@asobj.register
def asobj_structtree(obj: _Tree) -> StructTree:
    root = StructElement(
        type="StructTreeRoot",
        children=[asobj(el) for el in obj],
    )
    return StructTree(root=root)


@asobj.register
def asobj_document(pdf: _Document, exclude: Set[str] = set()) -> Document:
    doc = Document(
        pdf_version=pdf.pdf_version,
        is_printable=pdf.is_printable,
        is_modifiable=pdf.is_modifiable,
        is_extractable=pdf.is_extractable,
    )
    if pdf.encryption is not None:
        ids, encrypt = pdf.encryption
        a, b = ids
        doc["encryption"] = Encryption(ids=asobj(ids), encrypt=asobj(encrypt))
    if "pages" not in exclude:
        doc["pages"] = [asobj(page) for page in pdf.pages]
    if "objects" not in exclude:
        doc["objects"] = [asobj(obj) for obj in pdf.objects]
    if "outlines" not in exclude:
        if pdf.outline is not None:
            doc["outlines"] = asobj(pdf.outline)
        dests = asobj(pdf.destinations)
        if dests:
            doc["destinations"] = dests
    if "structure" not in exclude and pdf.structure is not None:
        doc["structure"] = asobj(pdf.structure)

    return doc



"""
Test PDF types and data structures.
"""

import pytest
from playa.pdftypes import (
    LIT,
    LITERAL_CRYPT,
    ContentStream,
    ObjRef,
    bool_value,
    decipher_all,
    decompress_corrupted,
    dict_value,
    float_value,
    keyword_name,
    list_value,
    matrix_value,
    num_value,
    point_value,
    rect_value,
    resolve1,
    resolve_all,
    str_value,
)
from playa.runlength import rldecode
from playa.worker import _ref_document


def test_rle():
    large_white_image_encoded = bytes([129, 255] * (3 * 3000 * 4000 // 128))
    _ = rldecode(large_white_image_encoded)


def test_resolve_all():
    """See if `resolve_all` will really `resolve` them `all`."""

    # Use a mock document, it just needs to suppot __getitem__
    class MockDoc(dict):
        pass

    mockdoc = MockDoc({42: b"hello"})
    mockdoc[41] = ObjRef(_ref_document(mockdoc), 42)
    mockdoc[40] = ObjRef(_ref_document(mockdoc), 41)
    assert mockdoc[41].resolve() == b"hello"
    assert resolve1(mockdoc[41]) == b"hello"
    assert mockdoc[40].resolve() == mockdoc[41]
    assert resolve_all(mockdoc[40]) == b"hello"
    mockdoc[39] = [mockdoc[40], mockdoc[41]]
    assert resolve_all(mockdoc[39]) == [b"hello", b"hello"]
    mockdoc[38] = [b"hello", ObjRef(_ref_document(mockdoc), 38)]
    # This resolves the *list*, not the indirect object, so its second
    # element will get expanded once into a new list.
    ouf = resolve_all(mockdoc[38])
    assert ouf[0] == b"hello"
    assert ouf[1][1] is mockdoc[38]
    # Whereas in this case we are expanding the reference itself.
    fou = resolve_all(mockdoc[38][1])
    assert fou[1] is mockdoc[38]
    # Likewise here, we have to dig a bit to see the circular
    # reference.  Your best option is not to use resolve_all ;-)
    mockdoc[30] = [b"hello", ObjRef(_ref_document(mockdoc), 31)]
    mockdoc[31] = [b"hello", ObjRef(_ref_document(mockdoc), 30)]
    bof = resolve_all(mockdoc[30])
    assert bof[1][1][1] is mockdoc[31]
    fob = resolve_all(mockdoc[30][1])
    assert fob[1][1] is mockdoc[31]


def test_errors() -> None:
    """Verify that we get various errors in pdftypes functions."""
    with pytest.raises(TypeError):
        keyword_name("NOT A KEYWORD DO NOT EAT")
    assert ObjRef(None, 1) == ObjRef(None, 1)
    assert ObjRef(None, 1) != ObjRef(123, 1)
    assert ObjRef(None, 1).resolve(123) == 123
    assert decipher_all(lambda *args: b"SOMETHING", 1, 0, b"") == b""
    with pytest.raises(TypeError):
        bool_value(b"Norwegian Blue")
    with pytest.raises(TypeError):
        float_value(b"Romanus eunt domus")
    float_value(123.456)  # just for the coverage
    with pytest.raises(TypeError):
        num_value(b"NaN")
    with pytest.raises(TypeError):
        str_value(42)
    with pytest.raises(TypeError):
        list_value({"not": "a list"})
    with pytest.raises(TypeError):
        dict_value(["not", "a dict"])
    with pytest.raises(ValueError):
        point_value([1, 2, 3])
    with pytest.raises(TypeError):
        point_value([32, "skidoo"])
    with pytest.raises(ValueError):
        rect_value([1, 2, 3])
    with pytest.raises(TypeError):
        rect_value([32, "skidoo", 4, 5])
    with pytest.raises(ValueError):
        matrix_value([1, 2, 3])
    with pytest.raises(TypeError):
        matrix_value([32, "skidoo", 4, 5, 6, 7])  # type: ignore[arg-type]
    broken_image = ContentStream(
        {"ColorSpace": LIT("NotAColorSpace")},
        rawdata=b"IRRELEVANT!",
    )
    with pytest.raises(ValueError):
        _ = broken_image.colorspace


def test_filters() -> None:
    """Exercise various filter types in streams"""
    stream = ContentStream({"Filter": [LITERAL_CRYPT]}, rawdata=b"")
    with pytest.raises(NotImplementedError):
        stream.decode()
    stream = ContentStream({"Filter": [LIT("FlateDecode")]}, rawdata=b"TOTAL NONSENSE")
    with pytest.raises(ValueError):
        stream.decode(strict=True)
    assert stream.buffer == b""
    stream = ContentStream(
        {"Filter": [LIT("LZWDecode")]}, rawdata=b"\x80\x0b\x60\x50\x22\x0c\x0c\x85\x01"
    )
    assert stream.buffer == b"\x2d\x2d\x2d\x2d\x2d\x41\x2d\x2d\x2d\x42"
    stream = ContentStream(
        {"Filter": [LIT("RunLengthDecode")]},
        rawdata=b"\x05123456\xfa7\x04abcde\x80junk",
    )
    assert stream.buffer == b"1234567777777abcde"
    stream = ContentStream(
        {"Filter": LIT("FlateDecode"), "DecodeParms": {"Predictor": 3}}, rawdata=b""
    )
    with pytest.raises(NotImplementedError):
        stream.decode()


def test_decompress_corrupted(caplog) -> None:
    """Verify that we can recover from missing CRC, truncated, or
    corrupted flate streams."""
    rawdata = (
        b'x\x9c\x0b\xf1\x0fq\xf4Qp\xf4sQ\x08\r\tq\rR\xf0\xf3\xf7\x0bv\x05"\x00R'
        b"\xe8\x06\xb5"
    )
    # Verify that decompress_corrupted works even on uncorrupted streams
    assert decompress_corrupted(rawdata) == b"TOTAL AND UTTER NONSENSE"
    # Or if the CRC is partially missing
    assert decompress_corrupted(rawdata[:-1]) == b"TOTAL AND UTTER NONSENSE"
    # Verify that decode will fail if we remove the trailer in strict mode
    stream = ContentStream({"Filter": LIT("FlateDecode")}, rawdata=rawdata[:-5])
    with pytest.raises(ValueError):
        stream.decode(strict=True)
    # Remove the trailer in non-strict mode
    stream = ContentStream({"Filter": LIT("FlateDecode")}, rawdata=rawdata[:-5])
    assert stream.buffer == b"TOTAL AND UTTER NONSENSE"
    # Remove the CRC in non-strict mode
    stream = ContentStream({"Filter": LIT("FlateDecode")}, rawdata=rawdata[:-3])
    assert stream.buffer == b"TOTAL AND UTTER NONSENSE"
    # Truncate the stream in non-strict mode
    stream = ContentStream({"Filter": LIT("FlateDecode")}, rawdata=rawdata[:-8])
    caplog.clear()
    assert stream.buffer == b"TOTAL AND UTTER NON"
    assert "incomplete" in caplog.text
    # Remove some random bytes (no error as the data is still a valid
    # flate stream, just the CRC is wrong)
    stream = ContentStream(
        {"Filter": LIT("FlateDecode")},
        rawdata=(rawdata[:5] + rawdata[7:10] + rawdata[11:]),
    )
    assert stream.buffer == b"TOT AL UTTER NONSENSE"
    # Insert some random bytes, now we will lose data for real
    bogusdata = bytearray(rawdata)
    bogusdata[7:10] = b"HI!"
    stream = ContentStream({"Filter": LIT("FlateDecode")}, rawdata=bogusdata)
    caplog.clear()
    assert stream.buffer == b""
    assert "Data loss" in caplog.text
    # Test the adjustible buffer size
    assert decompress_corrupted(bogusdata) == b""
    assert decompress_corrupted(bogusdata, 1) == b"TOTAHdd"
    assert decompress_corrupted(bogusdata, 2) == b"TOTAHdd"
    assert decompress_corrupted(bogusdata, 4) == b"TOTAHdd"
    assert decompress_corrupted(bogusdata, 8) == b"TOTAH"
    assert decompress_corrupted(bogusdata, 16) == b""



from playa.data._asobj import asobj
from playa.data.content import (
    Color,
    ColorSpace,
    DashPattern,
    Glyph,
    GraphicState,
    Image,
    MarkedContent,
    Path,
    PathSegment,
    Tag,
    Text,
)
from playa.data.metadata import (
    Annotation,
    Destination,
    Document,
    Encryption,
    Font,
    IndirectObject,
    Outline,
    Page,
    Resources,
    StreamObject,
    StructElement,
    StructTree,
    XObject,
)

__all__ = [
    "Color",
    "ColorSpace",
    "DashPattern",
    "Glyph",
    "GraphicState",
    "Image",
    "MarkedContent",
    "Path",
    "PathSegment",
    "Tag",
    "Text",
    "Annotation",
    "Destination",
    "Document",
    "Encryption",
    "Font",
    "IndirectObject",
    "Outline",
    "Page",
    "Resources",
    "StreamObject",
    "StructElement",
    "StructTree",
    "XObject",
    "asobj",
]
VERSION = "1.0"



"""
PLAYA ain't a LAYout Analyzer... but it can help you get stuff
out of PDFs.

Basic usage:

    with playa.open(path) as pdf:
        for page in pdf.pages:
            print(f"page {page.label}:")
            for obj in page:
                print(f"    {obj.object_type} at {obj.bbox}")
                if obj.object_type == "text":
                    print(f"        chars: {obj.chars}")
"""

import builtins
from concurrent.futures import ProcessPoolExecutor
from multiprocessing.context import BaseContext
from os import PathLike
from typing import Union

from playa._version import __version__
from playa.color import Color, ColorSpace
from playa.content import (
    ContentObject,
    GraphicState,
    MarkedContent,
    PathSegment,
)
from playa.data import asobj
from playa.document import Document, PageList
from playa.exceptions import PDFEncryptionError, PDFException, PDFPasswordIncorrect
from playa.outline import Destination, Action
from playa.page import DeviceSpace, Page, Annotation
from playa.parser import Token
from playa.pdftypes import ContentStream, Matrix, ObjRef, Point, Rect
from playa.pdftypes import resolve1 as resolve
from playa.pdftypes import resolve_all
from playa.worker import _init_worker, _init_worker_buffer

__all__ = [
    "Document",
    "Page",
    "PageList",
    "DeviceSpace",
    "Color",
    "ColorSpace",
    "ContentObject",
    "ContentStream",
    "Destination",
    "Action",
    "Annotation",
    "GraphicState",
    "MarkedContent",
    "Matrix",
    "Point",
    "Rect",
    "PathSegment",
    "Token",
    "ObjRef",
    "PDFEncryptionError",
    "PDFException",
    "PDFPasswordIncorrect",
    "asobj",
    "resolve",
    "resolve_all",
    "__version__",
]


def open(
    path: Union[PathLike, str],
    *,
    password: str = "",
    space: DeviceSpace = "screen",
    max_workers: Union[int, None] = 1,
    mp_context: Union[BaseContext, None] = None,
) -> Document:
    """Open a PDF document from a path on the filesystem.

    Args:
        path: Path to the document to open.
        space: Device space to use ("screen" for screen-like
               coordinates, "page" for pdfminer.six-like coordinates, "default" for
               default user space with no rotation or translation)
        max_workers: Number of worker processes to use for parallel
                     processing of pages (if 1, no workers are spawned)
        mp_context: Multiprocessing context to use for worker
                    processes, see [Contexts and Start
                    Methods](https://docs.python.org/3/library/multiprocessing.html#contexts-and-start-methods)
                    for more information.
    """
    fp = builtins.open(path, "rb")
    pdf = Document(fp, password=password, space=space)
    pdf._fp = fp
    if max_workers is not None and max_workers <= 0:
        raise ValueError("max_workers must be None or greater than 0")
    pdf._ncpus = 1 if max_workers is None else max_workers
    if max_workers == 1:
        return pdf
    # Eagerly load a few things
    catalog = pdf.catalog
    xrefs = pdf.xrefs
    pages = pdf.pages
    pdf.pages._eager_load()
    pdf._pool = ProcessPoolExecutor(
        max_workers=max_workers,
        mp_context=mp_context,
        initializer=_init_worker,  # type: ignore[arg-type]
        initargs=(  # type: ignore[arg-type]
            id(pdf),
            path,
            password,
            {
                "space": space,
                "_catalog": catalog,
                "_xrefs": xrefs,
                "_pages": pages,
            },
        ),
    )
    return pdf


def parse(
    buffer: bytes,
    *,
    password: str = "",
    space: DeviceSpace = "screen",
    max_workers: Union[int, None] = 1,
    mp_context: Union[BaseContext, None] = None,
) -> Document:
    """Read a PDF document from binary data.

    Note: Potential slowness
        When using multiple processes, this results in the entire
        buffer being copied to the worker processes for the moment,
        which may cause some overhead.  It is preferable to use `open`
        on a filesystem path if possible, since that uses
        memory-mapped I/O.

    Args:
        buffer: Buffer containing PDF data.
        space: Device space to use ("screen" for screen-like
               coordinates, "page" for pdfminer.six-like coordinates, "default" for
               default user space with no rotation or translation)
        max_workers: Number of worker processes to use for parallel
                     processing of pages (if 1, no workers are spawned)
        mp_context: Multiprocessing context to use for worker
                    processes, see [Contexts and Start
                    Methods](https://docs.python.org/3/library/multiprocessing.html#contexts-and-start-methods)
                    for more information.

    """
    pdf = Document(buffer, password=password, space=space)
    if max_workers is not None and max_workers <= 0:
        raise ValueError("max_workers must be None or greater than 0")
    pdf._ncpus = 1 if max_workers is None else max_workers
    if max_workers == 1:
        return pdf
    # Eagerly load a few things
    catalog = pdf.catalog
    xrefs = pdf.xrefs
    pages = pdf.pages
    pdf.pages._eager_load()
    pdf._pool = ProcessPoolExecutor(
        max_workers=max_workers,
        mp_context=mp_context,
        initializer=_init_worker_buffer,  # type: ignore[arg-type]
        initargs=(  # type: ignore[arg-type]
            id(pdf),
            buffer,
            password,
            {
                "space": space,
                "_catalog": catalog,
                "_xrefs": xrefs,
                "_pages": pages,
            },
        ),
    )
    return pdf



"""
Test pdfminer.six replacement functionality.
"""

from playa.miner import (
    extract,
    LAParams,
    LTFigure,
    LTTextLine,
    LTComponent,
    make_path_segment,
)
from pdfminer.high_level import extract_pages as pdfminer_extract_pages
from tests.data import TESTDIR, CONTRIB, XFAILS, PDFMINER_BUGS
from pathlib import Path
import pytest
import re
from typing import Any


# For the moment test a restricted set of PDFs as we are not totally
# bug-compatible with pdfminer.six
TESTPDFS = [
    (TESTDIR / name)
    for name in [
        "core_font_encodings.pdf",
        "font-size-test.pdf",
        "hello_structure.pdf",
        "simple1.pdf",
        "simple2.pdf",
        #    "vertical_writing.pdf",
        "zen_of_python_corrupted.pdf",
    ]
]


def compare(playa_item: Any, pdfminer_item: Any) -> None:
    """Compare PLAYA and pdfminer.six accounting for some differences."""
    pvstr = str(playa_item)
    pmstr = str(pdfminer_item)
    # Remove (cid:N) as PLAYA does not do that
    pmstr = re.sub(r"\(cid:\d+\)", "", pmstr)
    assert pvstr == pmstr


@pytest.mark.parametrize("path", TESTPDFS, ids=str)
def test_extract(path: Path):
    if path.name in XFAILS:
        pytest.xfail("Expected failure: %s" % path.name)
    if path.name in PDFMINER_BUGS:
        pytest.xfail("Skipping pdfminer.six failure: %s" % path.name)
    for idx, (playa_ltpage, pdfminer_ltpage) in enumerate(
        zip(extract(path, LAParams()), pdfminer_extract_pages(path))
    ):
        # Otherwise pdfminer.six is just too darn slow
        if idx == 20:
            break
        for playa_item, pdfminer_item in zip(playa_ltpage, pdfminer_ltpage):
            compare(playa_item, pdfminer_item)


@pytest.mark.skipif(not CONTRIB.exists(), reason="contrib samples not present")
def test_serialization():
    """Ensure stuff is reserialized properly"""
    path = CONTRIB / "PSC_Station.pdf"
    pages = extract(path, LAParams())
    first = next(pages)
    for item in first:
        if isinstance(item, LTFigure):
            img = next(iter(item))
            # We have a image stream
            assert len(img.stream.buffer) == 52692
            assert img.colorspace.name == "ICCBased"
            # It has a colorspace, which has a stream as an indirect
            # object reference, which we can resolve
            icc = img.colorspace.spec[1].resolve()
            assert len(icc.buffer) == 3144
        # Probably we could test some other things too?


def test_make_path_segment() -> None:
    """Verify make_path_segment works"""
    assert make_path_segment("h", []) == ("h",)
    # NOTE: This is the bogus output type for bug compatibility with pdfminer.six
    assert make_path_segment("m", [(1, 2)]) == ("m", (1, 2))
    assert make_path_segment("l", [(3, 4)]) == ("l", (3, 4))
    assert make_path_segment("v", [(3, 4), (5, 6)]) == ("v", (3, 4), (5, 6))
    assert make_path_segment("y", [(3, 4), (5, 6)]) == ("y", (3, 4), (5, 6))
    assert make_path_segment("c", [(1, 2), (3, 4), (5, 6)]) == (
        "c",
        (1, 2),
        (3, 4),
        (5, 6),
    )
    with pytest.raises(ValueError):
        make_path_segment("h", [(1, 2)])
        make_path_segment("m", [(1, 2), (3, 4)])
        make_path_segment("l", [(1, 2), (3, 4)])
        make_path_segment("v", [(1, 2)])
        make_path_segment("y", [(1, 2)])
        make_path_segment("c", [(1, 2)])


def test_items_are_hashable_and_serializable() -> None:
    """Verify that we can hash and serialize LTThingies"""
    import pickle

    assert hash(LTTextLine(0.3))
    assert hash(LTComponent((1, 2, 3, 4), ()))
    data = pickle.dumps(LTComponent((1, 2, 3, 4), ()))
    assert hash(pickle.loads(data))


if __name__ == "__main__":
    test_extract()



import logging
import struct
from collections import deque
from io import BytesIO
from typing import BinaryIO, Deque, Dict, List, Iterator, Tuple, Union

from playa.cmapdb import ToUnicodeMap
from playa.parser import Lexer, Token, KWD, PSLiteral
from playa.utils import nunpack

log = logging.getLogger(__name__)


class CFFFontProgram:
    NIBBLES = (
        "0",
        "1",
        "2",
        "3",
        "4",
        "5",
        "6",
        "7",
        "8",
        "9",
        ".",
        "e",
        "e-",
        None,
        "-",
    )

    STANDARD_STRINGS = (
        ".notdef",
        "space",
        "exclam",
        "quotedbl",
        "numbersign",
        "dollar",
        "percent",
        "ampersand",
        "quoteright",
        "parenleft",
        "parenright",
        "asterisk",
        "plus",
        "comma",
        "hyphen",
        "period",
        "slash",
        "zero",
        "one",
        "two",
        "three",
        "four",
        "five",
        "six",
        "seven",
        "eight",
        "nine",
        "colon",
        "semicolon",
        "less",
        "equal",
        "greater",
        "question",
        "at",
        "A",
        "B",
        "C",
        "D",
        "E",
        "F",
        "G",
        "H",
        "I",
        "J",
        "K",
        "L",
        "M",
        "N",
        "O",
        "P",
        "Q",
        "R",
        "S",
        "T",
        "U",
        "V",
        "W",
        "X",
        "Y",
        "Z",
        "bracketleft",
        "backslash",
        "bracketright",
        "asciicircum",
        "underscore",
        "quoteleft",
        "a",
        "b",
        "c",
        "d",
        "e",
        "f",
        "g",
        "h",
        "i",
        "j",
        "k",
        "l",
        "m",
        "n",
        "o",
        "p",
        "q",
        "r",
        "s",
        "t",
        "u",
        "v",
        "w",
        "x",
        "y",
        "z",
        "braceleft",
        "bar",
        "braceright",
        "asciitilde",
        "exclamdown",
        "cent",
        "sterling",
        "fraction",
        "yen",
        "florin",
        "section",
        "currency",
        "quotesingle",
        "quotedblleft",
        "guillemotleft",
        "guilsinglleft",
        "guilsinglright",
        "fi",
        "fl",
        "endash",
        "dagger",
        "daggerdbl",
        "periodcentered",
        "paragraph",
        "bullet",
        "quotesinglbase",
        "quotedblbase",
        "quotedblright",
        "guillemotright",
        "ellipsis",
        "perthousand",
        "questiondown",
        "grave",
        "acute",
        "circumflex",
        "tilde",
        "macron",
        "breve",
        "dotaccent",
        "dieresis",
        "ring",
        "cedilla",
        "hungarumlaut",
        "ogonek",
        "caron",
        "emdash",
        "AE",
        "ordfeminine",
        "Lslash",
        "Oslash",
        "OE",
        "ordmasculine",
        "ae",
        "dotlessi",
        "lslash",
        "oslash",
        "oe",
        "germandbls",
        "onesuperior",
        "logicalnot",
        "mu",
        "trademark",
        "Eth",
        "onehalf",
        "plusminus",
        "Thorn",
        "onequarter",
        "divide",
        "brokenbar",
        "degree",
        "thorn",
        "threequarters",
        "twosuperior",
        "registered",
        "minus",
        "eth",
        "multiply",
        "threesuperior",
        "copyright",
        "Aacute",
        "Acircumflex",
        "Adieresis",
        "Agrave",
        "Aring",
        "Atilde",
        "Ccedilla",
        "Eacute",
        "Ecircumflex",
        "Edieresis",
        "Egrave",
        "Iacute",
        "Icircumflex",
        "Idieresis",
        "Igrave",
        "Ntilde",
        "Oacute",
        "Ocircumflex",
        "Odieresis",
        "Ograve",
        "Otilde",
        "Scaron",
        "Uacute",
        "Ucircumflex",
        "Udieresis",
        "Ugrave",
        "Yacute",
        "Ydieresis",
        "Zcaron",
        "aacute",
        "acircumflex",
        "adieresis",
        "agrave",
        "aring",
        "atilde",
        "ccedilla",
        "eacute",
        "ecircumflex",
        "edieresis",
        "egrave",
        "iacute",
        "icircumflex",
        "idieresis",
        "igrave",
        "ntilde",
        "oacute",
        "ocircumflex",
        "odieresis",
        "ograve",
        "otilde",
        "scaron",
        "uacute",
        "ucircumflex",
        "udieresis",
        "ugrave",
        "yacute",
        "ydieresis",
        "zcaron",
        "exclamsmall",
        "Hungarumlautsmall",
        "dollaroldstyle",
        "dollarsuperior",
        "ampersandsmall",
        "Acutesmall",
        "parenleftsuperior",
        "parenrightsuperior",
        "twodotenleader",
        "onedotenleader",
        "zerooldstyle",
        "oneoldstyle",
        "twooldstyle",
        "threeoldstyle",
        "fouroldstyle",
        "fiveoldstyle",
        "sixoldstyle",
        "sevenoldstyle",
        "eightoldstyle",
        "nineoldstyle",
        "commasuperior",
        "threequartersemdash",
        "periodsuperior",
        "questionsmall",
        "asuperior",
        "bsuperior",
        "centsuperior",
        "dsuperior",
        "esuperior",
        "isuperior",
        "lsuperior",
        "msuperior",
        "nsuperior",
        "osuperior",
        "rsuperior",
        "ssuperior",
        "tsuperior",
        "ff",
        "ffi",
        "ffl",
        "parenleftinferior",
        "parenrightinferior",
        "Circumflexsmall",
        "hyphensuperior",
        "Gravesmall",
        "Asmall",
        "Bsmall",
        "Csmall",
        "Dsmall",
        "Esmall",
        "Fsmall",
        "Gsmall",
        "Hsmall",
        "Ismall",
        "Jsmall",
        "Ksmall",
        "Lsmall",
        "Msmall",
        "Nsmall",
        "Osmall",
        "Psmall",
        "Qsmall",
        "Rsmall",
        "Ssmall",
        "Tsmall",
        "Usmall",
        "Vsmall",
        "Wsmall",
        "Xsmall",
        "Ysmall",
        "Zsmall",
        "colonmonetary",
        "onefitted",
        "rupiah",
        "Tildesmall",
        "exclamdownsmall",
        "centoldstyle",
        "Lslashsmall",
        "Scaronsmall",
        "Zcaronsmall",
        "Dieresissmall",
        "Brevesmall",
        "Caronsmall",
        "Dotaccentsmall",
        "Macronsmall",
        "figuredash",
        "hypheninferior",
        "Ogoneksmall",
        "Ringsmall",
        "Cedillasmall",
        "questiondownsmall",
        "oneeighth",
        "threeeighths",
        "fiveeighths",
        "seveneighths",
        "onethird",
        "twothirds",
        "zerosuperior",
        "foursuperior",
        "fivesuperior",
        "sixsuperior",
        "sevensuperior",
        "eightsuperior",
        "ninesuperior",
        "zeroinferior",
        "oneinferior",
        "twoinferior",
        "threeinferior",
        "fourinferior",
        "fiveinferior",
        "sixinferior",
        "seveninferior",
        "eightinferior",
        "nineinferior",
        "centinferior",
        "dollarinferior",
        "periodinferior",
        "commainferior",
        "Agravesmall",
        "Aacutesmall",
        "Acircumflexsmall",
        "Atildesmall",
        "Adieresissmall",
        "Aringsmall",
        "AEsmall",
        "Ccedillasmall",
        "Egravesmall",
        "Eacutesmall",
        "Ecircumflexsmall",
        "Edieresissmall",
        "Igravesmall",
        "Iacutesmall",
        "Icircumflexsmall",
        "Idieresissmall",
        "Ethsmall",
        "Ntildesmall",
        "Ogravesmall",
        "Oacutesmall",
        "Ocircumflexsmall",
        "Otildesmall",
        "Odieresissmall",
        "OEsmall",
        "Oslashsmall",
        "Ugravesmall",
        "Uacutesmall",
        "Ucircumflexsmall",
        "Udieresissmall",
        "Yacutesmall",
        "Thornsmall",
        "Ydieresissmall",
        "001.000",
        "001.001",
        "001.002",
        "001.003",
        "Black",
        "Bold",
        "Book",
        "Light",
        "Medium",
        "Regular",
        "Roman",
        "Semibold",
    )

    STANDARD_ENCODING = (
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        12,
        13,
        14,
        15,
        16,
        17,
        18,
        19,
        20,
        21,
        22,
        23,
        24,
        25,
        26,
        27,
        28,
        29,
        30,
        31,
        32,
        33,
        34,
        35,
        36,
        37,
        38,
        39,
        40,
        41,
        42,
        43,
        44,
        45,
        46,
        47,
        48,
        49,
        50,
        51,
        52,
        53,
        54,
        55,
        56,
        57,
        58,
        59,
        60,
        61,
        62,
        63,
        64,
        65,
        66,
        67,
        68,
        69,
        70,
        71,
        72,
        73,
        74,
        75,
        76,
        77,
        78,
        79,
        80,
        81,
        82,
        83,
        84,
        85,
        86,
        87,
        88,
        89,
        90,
        91,
        92,
        93,
        94,
        95,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        96,
        97,
        98,
        99,
        100,
        101,
        102,
        103,
        104,
        105,
        106,
        107,
        108,
        109,
        110,
        0,
        111,
        112,
        113,
        114,
        0,
        115,
        116,
        117,
        118,
        119,
        120,
        121,
        122,
        0,
        123,
        0,
        124,
        125,
        126,
        127,
        128,
        129,
        130,
        131,
        0,
        132,
        133,
        0,
        134,
        135,
        136,
        137,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        138,
        0,
        139,
        0,
        0,
        0,
        0,
        140,
        141,
        142,
        143,
        0,
        0,
        0,
        0,
        0,
        144,
        0,
        0,
        0,
        145,
        0,
        0,
        146,
        147,
        148,
        149,
        0,
        0,
        0,
        0,
    )
    EXPERT_ENCODING = (
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        229,
        230,
        0,
        231,
        232,
        233,
        234,
        235,
        236,
        237,
        238,
        13,
        14,
        15,
        99,
        239,
        240,
        241,
        242,
        243,
        244,
        245,
        246,
        247,
        248,
        27,
        28,
        249,
        250,
        251,
        252,
        0,
        253,
        254,
        255,
        256,
        257,
        0,
        0,
        0,
        258,
        0,
        0,
        259,
        260,
        261,
        262,
        0,
        0,
        263,
        264,
        265,
        0,
        266,
        109,
        110,
        267,
        268,
        269,
        0,
        270,
        271,
        272,
        273,
        274,
        275,
        276,
        277,
        278,
        279,
        280,
        281,
        282,
        283,
        284,
        285,
        286,
        287,
        288,
        289,
        290,
        291,
        292,
        293,
        294,
        295,
        296,
        297,
        298,
        299,
        300,
        301,
        302,
        303,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        304,
        305,
        306,
        0,
        0,
        307,
        308,
        309,
        310,
        311,
        0,
        312,
        0,
        0,
        313,
        0,
        0,
        314,
        315,
        0,
        0,
        316,
        317,
        318,
        0,
        0,
        0,
        158,
        155,
        163,
        319,
        320,
        321,
        322,
        323,
        324,
        325,
        0,
        0,
        326,
        150,
        164,
        169,
        327,
        328,
        329,
        330,
        331,
        332,
        333,
        334,
        335,
        336,
        337,
        338,
        339,
        340,
        341,
        342,
        343,
        344,
        345,
        346,
        347,
        348,
        349,
        350,
        351,
        352,
        353,
        354,
        355,
        356,
        357,
        358,
        359,
        360,
        361,
        362,
        363,
        364,
        365,
        366,
        367,
        368,
        369,
        370,
        371,
        372,
        373,
        374,
        375,
        376,
        377,
        378,
    )
    PREDEFINED_ENCODINGS = (STANDARD_ENCODING, EXPERT_ENCODING)

    ISOADOBE_CHARSET = (
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        12,
        13,
        14,
        15,
        16,
        17,
        18,
        19,
        20,
        21,
        22,
        23,
        24,
        25,
        26,
        27,
        28,
        29,
        30,
        31,
        32,
        33,
        34,
        35,
        36,
        37,
        38,
        39,
        40,
        41,
        42,
        43,
        44,
        45,
        46,
        47,
        48,
        49,
        50,
        51,
        52,
        53,
        54,
        55,
        56,
        57,
        58,
        59,
        60,
        61,
        62,
        63,
        64,
        65,
        66,
        67,
        68,
        69,
        70,
        71,
        72,
        73,
        74,
        75,
        76,
        77,
        78,
        79,
        80,
        81,
        82,
        83,
        84,
        85,
        86,
        87,
        88,
        89,
        90,
        91,
        92,
        93,
        94,
        95,
        96,
        97,
        98,
        99,
        100,
        101,
        102,
        103,
        104,
        105,
        106,
        107,
        108,
        109,
        110,
        111,
        112,
        113,
        114,
        115,
        116,
        117,
        118,
        119,
        120,
        121,
        122,
        123,
        124,
        125,
        126,
        127,
        128,
        129,
        130,
        131,
        132,
        133,
        134,
        135,
        136,
        137,
        138,
        139,
        140,
        141,
        142,
        143,
        144,
        145,
        146,
        147,
        148,
        149,
        150,
        151,
        152,
        153,
        154,
        155,
        156,
        157,
        158,
        159,
        160,
        161,
        162,
        163,
        164,
        165,
        166,
        167,
        168,
        169,
        170,
        171,
        172,
        173,
        174,
        175,
        176,
        177,
        178,
        179,
        180,
        181,
        182,
        183,
        184,
        185,
        186,
        187,
        188,
        189,
        190,
        191,
        192,
        193,
        194,
        195,
        196,
        197,
        198,
        199,
        200,
        201,
        202,
        203,
        204,
        205,
        206,
        207,
        208,
        209,
        210,
        211,
        212,
        213,
        214,
        215,
        216,
        217,
        218,
        219,
        220,
        221,
        222,
        223,
        224,
        225,
        226,
        227,
        228,
    )
    EXPERT_CHARSET = (
        1,
        229,
        230,
        231,
        232,
        233,
        234,
        235,
        236,
        237,
        238,
        13,
        14,
        15,
        99,
        239,
        240,
        241,
        242,
        243,
        244,
        245,
        246,
        247,
        248,
        27,
        28,
        249,
        250,
        251,
        252,
        253,
        254,
        255,
        256,
        257,
        258,
        259,
        260,
        261,
        262,
        263,
        264,
        265,
        266,
        109,
        110,
        267,
        268,
        269,
        270,
        271,
        272,
        273,
        274,
        275,
        276,
        277,
        278,
        279,
        280,
        281,
        282,
        283,
        284,
        285,
        286,
        287,
        288,
        289,
        290,
        291,
        292,
        293,
        294,
        295,
        296,
        297,
        298,
        299,
        300,
        301,
        302,
        303,
        304,
        305,
        306,
        307,
        308,
        309,
        310,
        311,
        312,
        313,
        314,
        315,
        316,
        317,
        318,
        158,
        155,
        163,
        319,
        320,
        321,
        322,
        323,
        324,
        325,
        326,
        150,
        164,
        169,
        327,
        328,
        329,
        330,
        331,
        332,
        333,
        334,
        335,
        336,
        337,
        338,
        339,
        340,
        341,
        342,
        343,
        344,
        345,
        346,
        347,
        348,
        349,
        350,
        351,
        352,
        353,
        354,
        355,
        356,
        357,
        358,
        359,
        360,
        361,
        362,
        363,
        364,
        365,
        366,
        367,
        368,
        369,
        370,
        371,
        372,
        373,
        374,
        375,
        376,
        377,
        378,
    )
    EXPERT_SUBSET_CHARSET = (
        1,
        231,
        232,
        235,
        236,
        237,
        238,
        13,
        14,
        15,
        99,
        239,
        240,
        241,
        242,
        243,
        244,
        245,
        246,
        247,
        248,
        27,
        28,
        249,
        250,
        251,
        253,
        254,
        255,
        256,
        257,
        258,
        259,
        260,
        261,
        262,
        263,
        264,
        265,
        266,
        109,
        110,
        267,
        268,
        269,
        270,
        272,
        300,
        301,
        302,
        305,
        314,
        315,
        158,
        155,
        163,
        320,
        321,
        322,
        323,
        324,
        325,
        326,
        150,
        164,
        169,
        327,
        328,
        329,
        330,
        331,
        332,
        333,
        334,
        335,
        336,
        337,
        338,
        339,
        340,
        341,
        342,
        343,
        344,
        345,
        346,
    )
    PREDEFINED_CHARSETS = (ISOADOBE_CHARSET, EXPERT_CHARSET, EXPERT_SUBSET_CHARSET)

    class INDEX:
        def __init__(self, fp: BinaryIO) -> None:
            self.fp = fp
            self.offsets: List[int] = []
            (count,) = struct.unpack(">H", self.fp.read(2))
            if count == 0:
                return
            (offsize,) = struct.unpack("B", self.fp.read(1))
            for i in range(count + 1):
                self.offsets.append(nunpack(self.fp.read(offsize)))
            self.base = self.fp.tell() - 1
            self.fp.seek(self.base + self.offsets[-1])

        def __repr__(self) -> str:
            return "<INDEX: size=%d>" % len(self)

        def __len__(self) -> int:
            return len(self.offsets) - 1

        def __getitem__(self, i: int) -> bytes:
            self.fp.seek(self.base + self.offsets[i])
            return self.fp.read(self.offsets[i + 1] - self.offsets[i])

        def __iter__(self) -> Iterator[bytes]:
            return iter(self[i] for i in range(len(self)))

    def __init__(self, name: str, fp: BinaryIO) -> None:
        self.name = name
        self.fp = fp
        # Header
        (_major, _minor, hdrsize, offsize) = struct.unpack("BBBB", self.fp.read(4))
        self.fp.read(hdrsize - 4)
        # Name INDEX
        self.name_index = self.INDEX(self.fp)
        # Top DICT INDEX
        self.dict_index = self.INDEX(self.fp)
        # String INDEX
        self.string_index = self.INDEX(self.fp)
        # Global Subr INDEX
        self.subr_index = self.INDEX(self.fp)
        # Top DICT DATA
        self.top_dict = self.getdict(self.dict_index[0])
        self.is_cidfont = (12, 30) in self.top_dict
        (charset_pos,) = self.top_dict.get(15, [0])
        (encoding_pos,) = self.top_dict.get(16, [0])
        (charstring_pos,) = self.top_dict.get(17, [0])
        # CharStrings
        self.fp.seek(int(charstring_pos))
        self.charstring = self.INDEX(self.fp)
        self.nglyphs = len(self.charstring)
        self._parse_charset(int(charset_pos))
        if not self.is_cidfont:
            self.name2gid = {self.getstr(sid): gid for gid, sid in self.gid2sid.items()}
            self._parse_encoding(int(encoding_pos))
            self.code2name = {
                code: self.getstr(self.gid2sid[gid])
                for code, gid in self.code2gid.items()
                if gid in self.gid2sid
            }
        else:
            self.cid2gid = {sid: gid for gid, sid in self.gid2sid.items()}

    def _parse_encoding(self, encoding_pos: int) -> None:
        # Encodings
        self.code2gid = {}
        if encoding_pos in (0, 1):
            for code, sid in enumerate(self.PREDEFINED_ENCODINGS[encoding_pos]):
                if gid := self.name2gid.get(self.getstr(sid)):
                    self.code2gid[code] = gid
            return
        self.fp.seek(encoding_pos)
        (format,) = self.fp.read(1)
        supp, format = format & 0x80, format & 0x7F
        self.encoding_format = format
        if format == 0:
            # Format 0
            (n,) = struct.unpack("B", self.fp.read(1))
            for gid, code in enumerate(
                struct.unpack("B" * n, self.fp.read(n)), start=1
            ):
                self.code2gid[code] = gid
        elif format == 1:
            # Format 1
            (n,) = struct.unpack("B", self.fp.read(1))
            gid = 1
            for i in range(n):
                (first, nleft) = struct.unpack("BB", self.fp.read(2))
                for code in range(first, first + nleft + 1):
                    self.code2gid[code] = gid
                    gid += 1
        else:
            raise ValueError("unsupported encoding format: %r" % format)
        if supp:
            (n,) = struct.unpack("B", self.fp.read(1))
            for i in range(n):
                code, sid = struct.unpack(">BH", self.fp.read(3))
                if gid := self.name2gid.get(self.getstr(sid)):
                    self.code2gid[code] = gid

    def _parse_charset(self, charset_pos: int) -> None:
        # Charsets
        self.gid2sid = {}
        if charset_pos in (0, 1, 2):
            if self.is_cidfont:
                raise ValueError("no predefined charsets for CID CFF fonts")
            for gid, sid in enumerate(self.PREDEFINED_CHARSETS[charset_pos], start=1):
                self.gid2sid[gid] = sid
            return
        self.fp.seek(charset_pos)
        (format,) = self.fp.read(1)
        self.charset_format = format
        if format == 0:
            # Format 0
            n = self.nglyphs - 1
            for gid, sid in enumerate(
                struct.unpack(">" + "H" * n, self.fp.read(2 * n)), start=1
            ):
                self.gid2sid[gid] = sid
        elif format in (1, 2):
            # Format 1 & 2
            range_f = ">HB" if format == 1 else ">HH"
            range_f_size = struct.calcsize(range_f)
            gid = 1
            while gid < self.nglyphs:
                (first, nleft) = struct.unpack(range_f, self.fp.read(range_f_size))
                for sid in range(first, first + nleft + 1):
                    self.gid2sid[gid] = sid
                    gid += 1
        else:
            raise ValueError("unsupported charset format: %r" % format)

    def getstr(self, sid: int) -> str:
        if sid < len(self.STANDARD_STRINGS):
            return self.STANDARD_STRINGS[sid]
        return self.string_index[sid - len(self.STANDARD_STRINGS)].decode("ascii")

    def getdict(
        self, data: bytes
    ) -> Dict[Union[Tuple[int, int], int], List[Union[float, int]]]:
        d: Dict[Union[Tuple[int, int], int], List[Union[float, int]]] = {}
        fp = BytesIO(data)
        stack: List[Union[float, int]] = []
        while 1:
            c = fp.read(1)
            if not c:
                break
            b0 = ord(c)
            if b0 <= 21:
                key = (12, ord(fp.read(1))) if b0 == 12 else b0
                d[key] = stack
                stack = []
                continue
            if b0 == 30:
                s = ""
                loop = True
                while loop:
                    b = ord(fp.read(1))
                    for n in (b >> 4, b & 15):
                        if n == 15:
                            loop = False
                        else:
                            nibble = self.NIBBLES[n]
                            assert nibble is not None
                            s += nibble
                value = float(s)
            elif b0 >= 32 and b0 <= 246:
                value = b0 - 139
            else:
                b1 = ord(fp.read(1))
                if b0 >= 247 and b0 <= 250:
                    value = ((b0 - 247) << 8) + b1 + 108
                elif b0 >= 251 and b0 <= 254:
                    value = -((b0 - 251) << 8) - b1 - 108
                else:
                    b2 = ord(fp.read(1))
                    if b1 >= 128:
                        b1 -= 256
                    if b0 == 28:
                        value = b1 << 8 | b2
                    else:
                        value = b1 << 24 | b2 << 16 | struct.unpack(">H", fp.read(2))[0]
            stack.append(value)
        return d


class TrueTypeFontProgram:
    """Read TrueType font programs to get Unicode mappings."""

    def __init__(self, name: str, fp: BinaryIO) -> None:
        self.name = name
        self.fp = fp
        self.tables: Dict[bytes, Tuple[int, int]] = {}
        self.fonttype = fp.read(4)
        try:
            (ntables, _1, _2, _3) = struct.unpack(">HHHH", fp.read(8))
            for _ in range(ntables):
                (name_bytes, tsum, offset, length) = struct.unpack(
                    ">4sLLL", fp.read(16)
                )
                self.tables[name_bytes] = (offset, length)
        except struct.error:
            # Do not fail if there are not enough bytes to read. Even for
            # corrupted PDFs we would like to get as much information as
            # possible, so continue.
            pass

    def create_tounicode(self) -> Union[ToUnicodeMap, None]:
        """Recreate a ToUnicode mapping from a TrueType font program."""
        if b"cmap" not in self.tables:
            log.debug("TrueType font program has no character mapping")
            return None
        (base_offset, length) = self.tables[b"cmap"]
        fp = self.fp
        fp.seek(base_offset)
        (version, nsubtables) = struct.unpack(">HH", fp.read(4))
        subtables: List[Tuple[int, int, int]] = []
        for i in range(nsubtables):
            subtables.append(struct.unpack(">HHL", fp.read(8)))
        char2gid: Dict[int, int] = {}
        # Only supports subtable type 0, 2 and 4.
        for platform_id, encoding_id, st_offset in subtables:
            # Skip non-Unicode cmaps.
            # https://docs.microsoft.com/en-us/typography/opentype/spec/cmap
            if not (platform_id == 0 or (platform_id == 3 and encoding_id in [1, 10])):
                continue
            fp.seek(base_offset + st_offset)
            (fmttype, fmtlen, fmtlang) = struct.unpack(">HHH", fp.read(6))
            if fmttype == 0:
                char2gid.update(enumerate(struct.unpack(">256B", fp.read(256))))
            elif fmttype == 2:
                subheaderkeys = struct.unpack(">256H", fp.read(512))
                firstbytes = [0] * 8192
                for i, k in enumerate(subheaderkeys):
                    firstbytes[k // 8] = i
                nhdrs = max(subheaderkeys) // 8 + 1
                hdrs: List[Tuple[int, int, int, int, int]] = []
                for i in range(nhdrs):
                    (firstcode, entcount, delta, offset) = struct.unpack(
                        ">HHhH", fp.read(8)
                    )
                    hdrs.append((i, firstcode, entcount, delta, fp.tell() - 2 + offset))
                for i, firstcode, entcount, delta, pos in hdrs:
                    if not entcount:
                        continue
                    first = firstcode + (firstbytes[i] << 8)
                    fp.seek(pos)
                    for c in range(entcount):
                        gid = struct.unpack(">H", fp.read(2))[0]
                        if gid:
                            gid += delta
                        char2gid[first + c] = gid
            elif fmttype == 4:
                (segcount, _1, _2, _3) = struct.unpack(">HHHH", fp.read(8))
                segcount //= 2
                ecs = struct.unpack(">%dH" % segcount, fp.read(2 * segcount))
                fp.read(2)
                scs = struct.unpack(">%dH" % segcount, fp.read(2 * segcount))
                idds = struct.unpack(">%dh" % segcount, fp.read(2 * segcount))
                pos = fp.tell()
                idrs = struct.unpack(">%dH" % segcount, fp.read(2 * segcount))
                for ec, sc, idd, idr in zip(ecs, scs, idds, idrs):
                    if idr:
                        fp.seek(pos + idr)
                        for c in range(sc, ec + 1):
                            b = struct.unpack(">H", fp.read(2))[0]
                            char2gid[c] = (b + idd) & 0xFFFF
                    else:
                        for c in range(sc, ec + 1):
                            char2gid[c] = (c + idd) & 0xFFFF
            else:
                # FIXME: support at least format 12 for non-BMP chars
                # (probably rare in real life since there should be a
                # ToUnicode mapping already)
                assert False, str(("Unhandled", fmttype))
        if not char2gid:
            log.debug("unicode mapping is empty")
            return None
        # Create unicode map - as noted above we don't yet support
        # Unicode outside the BMP, so this is 16-bit only.
        tounicode = ToUnicodeMap()
        tounicode.add_code_range(b"\x00\x00", b"\xff\xff")
        for char, gid in char2gid.items():
            tounicode.add_code2code(gid, char, 2)
        return tounicode


KEYWORD_BEGIN = KWD(b"begin")
KEYWORD_END = KWD(b"end")
KEYWORD_DEF = KWD(b"def")
KEYWORD_PUT = KWD(b"put")
KEYWORD_DICT = KWD(b"dict")
KEYWORD_ARRAY = KWD(b"array")
KEYWORD_READONLY = KWD(b"readonly")
KEYWORD_FOR = KWD(b"for")


class Type1FontHeaderParser:
    def __init__(self, data: bytes) -> None:
        self._lexer = Lexer(data)
        self._encoding: Dict[int, str] = {}
        self._tokq: Deque[Token] = deque([], 2)

    def get_encoding(self) -> Dict[int, str]:
        """Parse the font encoding.

        The Type1 font encoding maps character codes to character names. These
        character names could either be standard Adobe glyph names, or
        character names associated with custom CharStrings for this font. A
        CharString is a sequence of operations that describe how the character
        should be drawn. Currently, this function returns '' (empty string)
        for character names that are associated with a CharStrings.

        Reference: Adobe Systems Incorporated, Adobe Type 1 Font Format

        :returns mapping of character identifiers (cid's) to unicode characters
        """
        for _, tok in self._lexer:
            # Ignore anything that isn't INT NAME put
            if tok is KEYWORD_PUT:
                cid, name = self._tokq
                if isinstance(cid, int) and isinstance(name, PSLiteral):
                    self._encoding[cid] = name.name
            else:
                self._tokq.append(tok)
        return self._encoding



"""Python implementation of Arcfour encryption algorithm.
See https://en.wikipedia.org/wiki/RC4
This code is in the public domain.

"""

from typing import List, Sequence


class Arcfour:
    def __init__(self, key: Sequence[int]) -> None:
        s: List[int] = list(range(256))
        j = 0
        kpos = 0
        klen = len(key)
        for i in range(256):
            j = (j + s[i] + key[kpos]) & 0xFF
            (s[i], s[j]) = (s[j], s[i])
            kpos += 1
            if kpos == klen:
                kpos = 0
        self.s = s
        (self.i, self.j) = (0, 0)

    def process(self, data: bytes) -> bytes:
        (i, j) = (self.i, self.j)
        s = self.s
        r = bytearray()
        for c in data:
            i = (i + 1) & 0xFF
            j = (j + s[i]) & 0xFF
            (s[i], s[j]) = (s[j], s[i])
            k = s[(s[i] + s[j]) & 0xFF]
            r.append(c ^ k)
        (self.i, self.j) = (i, j)
        return bytes(r)
